package de.korap.json;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import com.fasterxml.jackson.databind.node.ObjectNode;

import de.korap.services.LemmaResponse;

public class TraversedPath {
	
	boolean
		insideQuery;
	int
		level;
	boolean
		okType; 
	int
		queryType; 			// LEMMA, etc.
	List<String>
		lemmata 	= null;	// the extracted lemmata.
	String
		lemma 		= null;	// currently found lemma.
	List<String>
		refs		= null;	// references to expansion lists returned from GLEMM.
							// same order than lemmata in lemmata[].
	JsonNode
		wrapNode	= null; // node under 'wrap' where properties of lemma term 
							// are stored, here called level 2.
	ObjectMapper
		objMapper	= null;
	int
		iRewrite	= 0;	// index of lemma currently rewritten (in lemmata).
	List<List<String>>
		expansionLists	= null;	// each element of this list is an expansion list
								// i.e. a list of wfs for a lemma.
	List<LemmaResponse>			// list of lemmaResponse corresponding to each
		lemRespList		= null;	// lemma of list lemmata.
								// todo: should replace List expansionLists.
	
	public static final int NONE		= 0; // undefined.
	public static final int LEMMA		= 1; // GLEMM lemma.
	public static final int REGEXPR		= 2; // COSMAS II's regex on word level.
	public static final int WILDCARD	= 3; // COSMAS II's wildcard expression.
	
	public TraversedPath(ObjectMapper obj)
	
		{
		objMapper	= obj;
		insideQuery = false;
		okType 		= false;
		level 		= 0;
		queryType	= NONE;
		}
	
	/*
	 * checkPath
	 * 
	 * notes:
	 * - to be called for each property of an object.
	 * - collects all lemma subqueries inside a query tree.
	 * Arguments:
	 * fieldName: currently processed field name.
	 * node		: JsonNode that is the value of fieldName.
	 * mode		: COLLECT or REWRITE.
	 * Returns:
	 * - List<String>lemmata as a list of lemmata.
	 * - true if wrap-object to be rewritten has been found.
	 * 04.04.20/FB
	 */
	
	boolean checkPath(final String fieldName, JsonNode node, final int mode, final PrintStream fout)
	
	{
	final String func = "checkPath";
	final boolean
		bLog 	= true; 	// for loging.
	boolean
		isFound = false;	// true, when wrap-obj to be rewritten has been found.
	
	if( insideQuery == false )
		{
		if( fieldName.equals("query") )
			{
			insideQuery = true;
			level = 1;
			okType = false;
			}
		return false;
		}
	
	if( level == 1 )
		{
		if( !okType && fieldName.equals("@type") )
			{
			if( node.isTextual() && node.asText().equals("koral:token") )
				{
				okType = true;
				}
			return false;
			}
		else if( okType && fieldName.equals("wrap") )
			{
			level 		= 2;
			okType 		= false; // reset for next level.
			queryType	= NONE;
			lemma 		= null;
			wrapNode	= node;
			if( bLog && fout != null )
				fout.printf("Debug: %s: wrap-Obj: '%s'\n", func, node.toString());
			}
		return false;
		} // level 1
	
	if( level == 2 )
		{
		if( fieldName.equals("@type") )
			{
			if( node.isTextual() && node.asText().equals("koral:term") )
				okType = true;
			else
				{ // reset, this is not the requested @type.
				level = 1;
				okType = false;
				}
			}
		else if( fieldName.equals("layer") ) 
			{
			if( node.isTextual() && node.asText().equals("lemma") )
				queryType = LEMMA;
			}
		else if( fieldName.equals("key") )
			{
			if( node.isTextual() )
				lemma = node.asText();
			}
		
		if( okType && queryType == LEMMA && lemma != null )
			{
			if( mode == JsonTraverse.COLLECT )
				{
				if( lemmata == null )
					lemmata = new ArrayList<String>();
				lemmata.add(lemma);
				
				if( bLog && fout != null )
					fout.printf("Debug: %s: found: lemma='%s' as koral:term.\n", func, lemma);
				}
			
			isFound = true; 
			
			// reset before searching the next lemma (i.e. @type); supposing we stay underneath node 'query'.
			level  = 1; 
			okType = false;
			}
		return isFound;
		} // level 2
	
	return false;
	} // checkPath

	/* rewriteObjNode
	 * notes: https://mkyong.com/java/jackson-tree-model-example/
	 * 08.04.20/FB
	 */
	
	public void rewriteObjNode(JsonNode node, final int mode, final PrintStream fout)
	
	{
	final String func = "rewriteObjNode";
	
	if( mode == JsonTraverse.REWRITE )
		{
		fout.printf("Debug: %s: rewriting for lemma '%s' ('%s') > '%s'...\n", func, 
					lemma, lemmata.get(iRewrite), refs.get(iRewrite));
		
		((ObjectNode)node).put("key",   "");
		((ObjectNode)node).put("layer", "orth");
		((ObjectNode)node).put("desc",  "&"+lemma);
		((ObjectNode)node).put("hide",  "key"); // TODO: must be ["key"] !
		((ObjectNode)node).put("ref",   refs.get(iRewrite)); // TODO: must be ["key"] !
		iRewrite++;
		}
	else // mode == REWRITE_EXP
		{
		fout.printf("Debug: %s: rewriting for lemma '%s'.\n", func, lemmata.get(iRewrite)); 
	
		// inserting expansion list from corresponding Lemma Response obj:
		ArrayNode
			array = objMapper.createArrayNode();
		LemmaResponse
			lemResp = lemRespList.get(iRewrite);
		
		for(int i=0; i<lemResp.listofWfs.size(); i++)
			array.add(lemResp.listofWfs.get(i));
		
		((ObjectNode)node).set("key",   array);	// expansion list as Json Node.
		((ObjectNode)node).put("layer", "orth");
		((ObjectNode)node).put("desc",  "&"+lemma);
		((ObjectNode)node).put("hide",  "key"); // TODO: must be ["key"] !
		iRewrite++;
		
		/* old version :
		ArrayNode
			array = objMapper.createArrayNode();
		List<String>
			curList = expansionLists.get(iRewrite);
		
		for(int i=0; i<curList.size(); i++)
			array.add(curList.get(i));
			
		((ObjectNode)node).set("key",   array);	// expansion list.
		((ObjectNode)node).put("layer", "orth");
		((ObjectNode)node).put("desc",  "&"+lemma);
		((ObjectNode)node).put("hide",  "key"); // TODO: must be ["key"] !
		iRewrite++;
		*/
		}

	} // rewriteObjNode
	
	/*
	 * getLemmata
	 * 
	 * returns list of lemmata collected and found in KoralQuery in String format.
	 * 
	 * 06.04.20/FB
	 */
	
	public String getLemmata()
	
	{
	return this.lemmata != null ? this.lemmata.toString() : "";
	} // getLemmata
	
	/* getLemmataCount
	 * 
	 * returns no. of lemmata found in KoralQuery.
	 * 07.04.20/FB
	 */
	
	public int getLemmaCount()
	
	{
	return this.lemmata != null ? this.lemmata.size() : 0;	
	} // getLemmataCount
	
	/* getLemma:
	 * returns lemma by its index i.
	 * returns null if index out of range.
	 * 07.04.20/FB
	 */
	
	public String getLemma(int i)
	
	{
	if( i < 0 || i > this.lemmata.size() )
		return null;
	
	return this.lemmata.get(i);
	} // getLemma
	
	/* setRef:
	 * - inserts a reference to a instanciation list into the
	 *   list refs by index i.
	 * - index i must correspond to lemma i in list lemmata.
	 * - this method ensures that list refs is allocated with as much positions
	 *   than the size of list lemmata.
	 * 07.04.20/FB
	 */
	 
	public void setRef(String ref)
	
	{
	if( this.refs == null )
		{
		int size = this.lemmata.size();
		this.refs = new ArrayList<String>(size);
		}
	
	this.refs.add(ref);
	} // setRef
	
	/* getRefsCount:
	 * returns no. of references stored in list this.refs.
	 * 07.04.20/FB
	 */
	
	public int getRefsCount()
	
	{
	return this.refs != null ? this.refs.size() : 0 ;	
	} // getRefsCount.
	
	/* getRefsasString:
	 * - returns a printable list of references together with the corresponding
	 *   lemmata.
	 * 07.04.20/FB
	 */
	
	public String getRefsasString()
	
	{
	int
		i, max; 
	
	if( this.lemmata == null || this.refs == null )
		return "";
	
	//max = this.lemmata.size() > this.refs.size() ? 
	//			this.lemmata.size() : this.refs.size();

	return this.refs.toString();
	} // getRefsasString
	
	/* addExpList
	 * 
	 * adds an Expansion List for lemma i.
	 * Notes:
	 * - add expList in same order as lemmata stored in list lemmata.
	 * - todo: will be replaced by addLemmaResponse().
	 * 08.04.20/FB
	 */
	
	public void addExpList(List expList)
	
	{
	if( expansionLists == null )
		{
		expansionLists = new ArrayList<List<String>>(lemmata.size());
		}
	
	expansionLists.add(expList);
	} // addExpList
	
	/*
	 * addLemmaResponse:
	 * - adds a LemmaResponse object for the correspondig lemma.
	 * 30.04.20/FB
	 * 
	 */
	
	public void addLemmaResponse(LemmaResponse lemResp)
	
	{
	if( this.lemRespList == null )
		this.lemRespList = new ArrayList<LemmaResponse>(lemmata.size());
	
	this.lemRespList.add(lemResp);
	} // addLemmaResponse
	
	/* rewriteSubQuery
	 * 
	 * Rules: Old			> 	New:
	 * 		  "key": "&Haus"	"key": "ref", or
	 * 		  "key": "&Haus"	"key": [Wf1, Wf2, ...].
	 * 		  "layer": "lemma"	"layer": "orth"
	 * 		  ""				"desc": "&Lemma" (neu)
	 * 		  ""				"hide": "key"    (neu)
	 * Arguments:
	 * node		: expected to be a "wrap"-object.
	 * 07.04.20/FB
	 */
	
	protected void rewriteSubQuery(JsonNode node, final String lemma, final PrintStream fout)
	
	{
	final String func = "rewriteSubQuery";
	final boolean
		bLog = true;
	
	if( bLog && fout != null)
		fout.printf("Debug: %s: rewrite '%s'.\n", func, node.toString());
	
	if( node.isObject() )
		{
	    Iterator<String> 
	    	fieldNames = node.fieldNames();
	    
		while(fieldNames.hasNext()) 
			{
		    String 
		    	fieldName 	= fieldNames.next();
		    JsonNode 
		    	fieldValue = node.get(fieldName);
		    
		    if( fieldValue.isTextual() == false )
		    	continue;
		    
		    // path search under 'query':
		    if( fieldName.equals("key") )
			    {
			    if( bLog && fout != null )
			    	fout.printf("Debug: %s: replacing %s='%s' by %s='%s'.\n", 
				    			func, fieldName, fieldValue.asText(),
				    			fieldName, "[...]");

		    	//JsonNode newNode = objMapper.createObjectNode();
			    //((ObjectNode)node).put("newkey", val);
			    ((ObjectNode)node).put("key", "[...]");
			    
			    if( bLog && fout != null )
			    	fout.printf("Debug: %s: %s='%s' (after replacement).\n", func, fieldName, node.get("key"));
			    }
		    else if( fieldName.equals("layer") )
			    {
			    	
			    }
		    
			} // while

		if( bLog && fout != null )
			fout.printf("Debug: %s: after rewriting, wrap-Obj='%s'.\n", func, node.toString());

		} // object
	
	} // rewriteSubQuery
	
	
} // class TraversedPath

